from pangocffi.ffi_instance_builder import FFIInstanceBuilder
from cffi import FFI


def test_ffi_instance_builder():
    ffi_builder = FFIInstanceBuilder('test')
    ffi = ffi_builder.generate()
    assert isinstance(ffi, FFI)



from pangocffi import Layout, ffi
from .context_creator import ContextCreator
import unittest


class TestItemWithContext(unittest.TestCase):

    def setUp(self):
        self.pango_context = ContextCreator.get_pango_context()

    def tearDown(self):
        self.pango_context = None
        ContextCreator.free()

    def test_item_properties(self):
        layout = Layout(self.pango_context)
        layout.set_text('Hi from Παν語')
        layout_iter = layout.get_iter()

        layout_run = layout_iter.get_run()
        layout_run_item = layout_run.item

        assert layout_run_item.length == 8
        assert layout_run_item.offset == 0
        assert layout_run_item.num_chars == 8
        assert isinstance(layout_run_item.get_pointer(), ffi.CData)

        assert layout_iter.next_run()
        layout_run = layout_iter.get_run()
        layout_run_item = layout_run.item

        assert layout_run_item.length == 6
        assert layout_run_item.offset == 8
        assert layout_run_item.num_chars == 3
        assert isinstance(layout_run_item.get_pointer(), ffi.CData)



from . import pango, ffi


def pango_version() -> int:
    """Return the pango version number as a single integer, such as 14204 for
    ``1.42.4``.
    Major, minor and micro versions are "worth" 10000, 100 and 1 respectively.

    Can be useful as a guard for method not available in older pango versions::

        if pango_version() >= 11000:
            ...

    """
    return pango.pango_version()


def pango_version_string() -> str:
    """Return the pango version number as a string, such as ``1.42.4``."""
    return ffi.string(pango.pango_version_string()).decode('ascii')



"""
    pangocffi.ffi_build
    ~~~~~~~~~~~~~~~~~~~

    Build the cffi bindings for pangocffi

"""

import sys
from pathlib import Path
import importlib.util

sys.path.append(str(Path(__file__).parent))

# Create an empty _generated folder if needed
(Path(__file__).parent / '_generated').mkdir(exist_ok=True)

# Because we can't directly load the instance builder (it would run
# ``__init__.py`` for any module import) we have to do this dubious import.
spec = importlib.util.spec_from_file_location(
    'ffi_instance_builder',
    str(Path(__file__).parent / 'ffi_instance_builder.py')
)
ffi_instance_builder = importlib.util.module_from_spec(spec)
spec.loader.exec_module(ffi_instance_builder)

# Generate the bindings
ffiBuilder = ffi_instance_builder.FFIInstanceBuilder(
    source='pangocffi._generated.ffi'
)
ffi = ffiBuilder.generate()

if __name__ == '__main__':
    ffi.compile()



from typing import Optional
from . import ffi


class Rectangle:
    """
    The :class:`Rectangle` structure represents a rectangle. It is
    frequently used to represent the logical or ink extents of a single glyph
    or section of text.
    """
    def __init__(
            self,
            pointer: Optional[ffi.CData] = None
    ):
        if pointer is None:
            self.pointer = ffi.new("PangoRectangle *")
        else:
            self.pointer = pointer

    def get_pointer(self) -> ffi.CData:
        """
        Returns a pointer to the :class:`Rectangle`.

        :return:
            a pointer to ``PangoRectangle``
        """
        return self.pointer

    @staticmethod
    def from_pointer(pointer: ffi.CData) -> 'Rectangle':
        """
        Returns an instance of a :class:`Rectangle` from a pointer.

        :return:
            the :class:`Rectangle`.
        """
        return Rectangle(
            pointer=pointer
        )

    @property
    def x(self) -> int:
        """
        :return:
            X coordinate of the left side of the rectangle.
        :type: int
        """
        return self.pointer.x

    @property
    def y(self) -> int:
        """
        :return:
            Y coordinate of the the top side of the rectangle.
        :type: int
        """
        return self.pointer.y

    @property
    def width(self) -> int:
        """
        :return:
            width of the rectangle.
        :type: int
        """
        return self.pointer.width

    @property
    def height(self) -> int:
        """
        :return:
            height of the rectangle.
        :type: int
        """
        return self.pointer.height



import pangocffi


def test_units_to_double():
    assert pangocffi.units_to_double(0) == 0
    assert pangocffi.units_to_double(123) == 0.1201171875


def test_context_properties():
    assert pangocffi.units_from_double(0) == 0
    assert pangocffi.units_from_double(0.1201171875) == 123



from typing import Optional, Union
from . import pango, ffi
from . import Style, Variant, Weight, Stretch, Gravity


class FontDescription(object):
    """
    The :class:`FontDescription` represents the description of an ideal font.
    These structures are used both to list what fonts are available on the
    system and also for specifying the characteristics of a font to load.
    """

    def __init__(self):
        """
        Creates a new font description structure with all fields unset.

        :param:
            a :class:`FontDescription`
        """
        self._init_pointer(pango.pango_font_description_new())

    def _init_pointer(self, pointer: ffi.CData):
        self._pointer = pointer
        # self._pointer = ffi.gc(pointer, pango.pango_font_description_free)

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to the font description

        :return:
            the pointer to the font description.
        """
        return self._pointer

    @classmethod
    def from_pointer(cls, pointer: ffi.CData) -> 'FontDescription':
        """
        Instantiates a :class:`FontDescription` from a pointer.

        :return:
            the font description.
        """
        if pointer == ffi.NULL:
            raise ValueError('Null pointer')
        self = object.__new__(cls)
        cls._init_pointer(self, pointer)
        return self

    def __eq__(self, other):
        if isinstance(other, FontDescription):
            return self.get_pointer() == other.get_pointer()
        return NotImplemented

    def set_family(self, family: str) -> None:
        """
        Sets the family name field of a font description. The family name
        represents a family of related font styles, and will resolve to a
        particular :class:`FontFamily`. In some uses of
        :class:`FontDescription`, it is also possible to use a comma separated
        list of family names for this field.

        :param family:
            a string representing the family name.
        """
        family_pointer = ffi.new('char[]', family.encode('utf8'))
        pango.pango_font_description_set_family(self._pointer, family_pointer)

    def get_family(self) -> Optional[str]:
        """
        Returns the family name field of a font description.

        :return:
            the family name field for the font description, or ``None`` if not
            previously set.
        """
        family_pointer = pango.pango_font_description_get_family(self._pointer)
        if family_pointer == ffi.NULL:
            return None
        return ffi.string(family_pointer).decode()

    def set_style(self, style: Style) -> None:
        """
        Sets the style field of a :class:`FontDescription`. The :class:`Style`
        enumeration describes whether the font is slanted and the manner in
        which it is slanted.

        Most fonts will either have a italic style or an oblique style, but not
        both, and font matching in Pango will match italic specifications with
        oblique fonts and vice-versa if an exact match is not found.

        :param style:
            the style for the font description
        """
        pango.pango_font_description_set_style(self._pointer, style.value)

    def get_style(self) -> Style:
        """
        Returns the style field of a :class:`FontDescription`.

        :return:
             the style field for the font description.
        """
        return Style(pango.pango_font_description_get_style(self._pointer))

    def set_variant(self, variant: Variant) -> None:
        """
        Sets the variant field of a font description. The :class:`Variant`
        enumeration describes whether the font is normal or small caps.

        :param variant:
            the variant type for the font description.
        """
        pango.pango_font_description_set_variant(self._pointer, variant.value)

    def get_variant(self) -> Variant:
        """
        Returns the variant field of a :class:`FontDescription`.

        :return:
            the variant field for the font description.
        """
        return Variant(pango.pango_font_description_get_variant(self._pointer))

    def set_weight(self, weight: Union[Weight, int]) -> None:
        """
        Sets the weight field of a font description. The weight field specifies
        how bold or light the font should be. In addition to the values of the
        :class:`Weight` enumeration, other intermediate numeric values are
        possible.

        :param weight:

        """
        if isinstance(weight, Weight):
            pango.pango_font_description_set_weight(
                self._pointer,
                weight.value
            )
        else:
            pango.pango_font_description_set_weight(self._pointer, weight)

    def get_weight(self) -> int:
        """
        Returns the weight field of a font description.

        :return:
            the weight field for the font description.
        """
        return pango.pango_font_description_get_weight(self._pointer)

    def set_stretch(self, stretch: Stretch) -> None:
        """
        Sets the stretch field of a font description. The stretch field
        specifies how narrow or wide the font should be.

        :param stretch:
             the stretch for the font description
        """
        pango.pango_font_description_set_stretch(self._pointer, stretch.value)

    def get_stretch(self) -> Stretch:
        """
        Returns the stretch field of a font description.

        :return:
            the stretch field for the font description.
        """
        return Stretch(pango.pango_font_description_get_stretch(self._pointer))

    def set_size(self, size: int) -> None:
        """
        Sets the size field of a font description in fractional points. This is
        mutually exclusive with :meth:`set_absolute_size()`.

        :param size:
             the size of the font in points, scaled by ``PANGO_SCALE``. (That
             is, a size value of 10 * ``PANGO_SCALE`` is a 10 point font. The
             conversion factor between points and device units depends on
             system configuration and the output device. For screen display, a
             logical DPI of 96 is common, in which case a 10 point font
             corresponds to a 10 * (96 / 72) = 13.3 pixel font. Use
             :meth:`set_absolute_size()` if you need a particular size in
             device units.
        """
        pango.pango_font_description_set_size(self._pointer, size)

    def get_size(self) -> int:
        """
        Returns the size field of a font description.

        :return:
            the size field for the font description in points or device units.
            You must call :meth:`get_size_is_absolute()` to find out which is
            the case. Returns 0 if the size field has not previously been set
            or it has been set to 0 explicitly.
        """
        return pango.pango_font_description_get_size(self._pointer)

    def set_absolute_size(self, size: float) -> None:
        """
        Sets the size field of a font description, in device units. This is
        mutually exclusive with :meth:`set_size()` which sets the font size in
        points.

        :param size:
             the new size, in Pango units. There are ``PANGO_SCALE`` Pango
             units in one device unit. For an output backend where a device
             unit is a pixel, a size value of 10 * ``PANGO_SCALE`` gives a 10
             pixel font.
        """
        pango.pango_font_description_set_absolute_size(self._pointer, size)

    def get_size_is_absolute(self) -> bool:
        """
        Determines whether the size of the font is in points (not absolute) or
        device units (absolute). See :meth:`set_size()` and
        :meth:`set_absolute_size()`.

        :return:
            whether the size for the font description is in points or device
            units.
        """
        return pango.pango_font_description_get_size_is_absolute(self._pointer)

    def set_gravity(self, gravity: Gravity) -> None:
        """
        Sets the gravity field of a font description. The gravity field
        specifies how the glyphs should be rotated. If gravity is
        ``Gravity.AUTO``, this actually unsets the gravity mask on the font
        description.

        :param gravity:
             the gravity for the font description.
        """
        pango.pango_font_description_set_gravity(self._pointer, gravity.value)

    def get_gravity(self) -> Gravity:
        """
        Returns the gravity field of a font description.

        :return:
            the gravity field for the font description.
        """
        return Gravity(pango.pango_font_description_get_gravity(self._pointer))



import pangocffi


def test_pango_version():
    version_code = pangocffi.pango_version()
    assert isinstance(version_code, int)


def test_pango_version_string():
    version_code_string = pangocffi.pango_version_string()
    assert isinstance(version_code_string, str)



import pangocffi
from pangocffi import Context
from pangocffi.ffi_build import ffi as ffi_builder
from cffi import FFI
import ctypes


def _dlopen(ffi, *names):
    """Try various names for the same library, for different platforms."""
    for name in names:
        for lib_name in (name, 'lib' + name):
            try:
                path = ctypes.util.find_library(lib_name)
                lib = ffi.dlopen(path or lib_name)
                if lib:
                    return lib
            except OSError:
                pass
    raise OSError("dlopen() failed to load a library: %s" % ' / '.join(names))


class ContextCreator(object):
    cairo = None
    cairo_surface_t = None
    cairo_t = None
    pango_context = None

    @staticmethod
    def get_pango_context() -> Context:
        ffi = FFI()
        ffi.include(ffi_builder)
        ffi.cdef('''
            /* Cairo */
            typedef void cairo_t;
            typedef struct _cairo_surface cairo_surface_t;

            typedef enum _cairo_status {
                CAIRO_STATUS_SUCCESS = 0,

                CAIRO_STATUS_NO_MEMORY,
                CAIRO_STATUS_INVALID_RESTORE,
                CAIRO_STATUS_INVALID_POP_GROUP,
                CAIRO_STATUS_NO_CURRENT_POINT,
                CAIRO_STATUS_INVALID_MATRIX,
                CAIRO_STATUS_INVALID_STATUS,
                CAIRO_STATUS_NULL_POINTER,
                CAIRO_STATUS_INVALID_STRING,
                CAIRO_STATUS_INVALID_PATH_DATA,
                CAIRO_STATUS_READ_ERROR,
                CAIRO_STATUS_WRITE_ERROR,
                CAIRO_STATUS_SURFACE_FINISHED,
                CAIRO_STATUS_SURFACE_TYPE_MISMATCH,
                CAIRO_STATUS_PATTERN_TYPE_MISMATCH,
                CAIRO_STATUS_INVALID_CONTENT,
                CAIRO_STATUS_INVALID_FORMAT,
                CAIRO_STATUS_INVALID_VISUAL,
                CAIRO_STATUS_FILE_NOT_FOUND,
                CAIRO_STATUS_INVALID_DASH,
                CAIRO_STATUS_INVALID_DSC_COMMENT,
                CAIRO_STATUS_INVALID_INDEX,
                CAIRO_STATUS_CLIP_NOT_REPRESENTABLE,
                CAIRO_STATUS_TEMP_FILE_ERROR,
                CAIRO_STATUS_INVALID_STRIDE,
                CAIRO_STATUS_FONT_TYPE_MISMATCH,
                CAIRO_STATUS_USER_FONT_IMMUTABLE,
                CAIRO_STATUS_USER_FONT_ERROR,
                CAIRO_STATUS_NEGATIVE_COUNT,
                CAIRO_STATUS_INVALID_CLUSTERS,
                CAIRO_STATUS_INVALID_SLANT,
                CAIRO_STATUS_INVALID_WEIGHT,
                CAIRO_STATUS_INVALID_SIZE,
                CAIRO_STATUS_USER_FONT_NOT_IMPLEMENTED,
                CAIRO_STATUS_DEVICE_TYPE_MISMATCH,
                CAIRO_STATUS_DEVICE_ERROR,
                CAIRO_STATUS_INVALID_MESH_CONSTRUCTION,
                CAIRO_STATUS_DEVICE_FINISHED,
                CAIRO_STATUS_JBIG2_GLOBAL_MISSING,
                CAIRO_STATUS_PNG_ERROR,
                CAIRO_STATUS_FREETYPE_ERROR,
                CAIRO_STATUS_WIN32_GDI_ERROR,
                CAIRO_STATUS_TAG_ERROR,

                CAIRO_STATUS_LAST_STATUS
            } cairo_status_t;

            typedef cairo_status_t (*cairo_write_func_t) (
                void * closure,
                const unsigned char *data,
                unsigned int length
            );
            cairo_surface_t * cairo_pdf_surface_create_for_stream (
                cairo_write_func_t write_func,
                void *closure,
                double width_in_points,
                double height_in_points
            );
            void cairo_surface_destroy (cairo_surface_t *surface);
            cairo_t * cairo_create (cairo_surface_t *target);
            void cairo_destroy (cairo_t *cr);

            PangoContext * pango_cairo_create_context (cairo_t *cr);
        ''')
        ffi.set_source('pangocffi._generated.ffi', None)
        cairo = _dlopen(ffi, 'cairo', 'cairo-2', 'cairo-gobject-2',
                        'cairo.so.2')
        pangocairo = _dlopen(ffi, 'pangocairo-1.0', 'pangocairo-1.0-0')

        cairo_surface_t = cairo.cairo_pdf_surface_create_for_stream(
            ffi.NULL,
            ffi.NULL,
            10,
            10
        )
        cairo_t = cairo.cairo_create(cairo_surface_t)

        pango_pointer = pangocairo.pango_cairo_create_context(cairo_t)
        pango_pointer = pangocffi.ffi.cast('PangoContext *', pango_pointer)
        pango_pointer = pangocffi.ffi.gc(
            pango_pointer,
            pangocffi.gobject.g_object_unref
        )

        ContextCreator.cairo = cairo
        ContextCreator.cairo_surface_t = cairo_surface_t
        ContextCreator.cairo_t = cairo_t
        ContextCreator.pango_context = Context.from_pointer(pango_pointer)

        return ContextCreator.pango_context

    @staticmethod
    def free():
        ContextCreator.pango_context = None
        ContextCreator.cairo.cairo_destroy(ContextCreator.cairo_t)
        ContextCreator.cairo.cairo_surface_destroy(
            ContextCreator.cairo_surface_t
        )



from . import GlyphItem


class LayoutRun(GlyphItem):
    """
    The :class:`LayoutRun` structure represents a single run within a
    :class:`LayoutLine`; it is simply an alternate name for :class:`GlyphItem`.
    See the :class:`GlyphItem` docs for details on the fields.
    """
    pass



from pangocffi import Layout, LayoutRun, Item, ffi
from .context_creator import ContextCreator
import unittest


class TestLayoutRunWithContext(unittest.TestCase):

    def setUp(self):
        self.pango_context = ContextCreator.get_pango_context()

    def tearDown(self):
        self.pango_context = None
        ContextCreator.free()

    def test_layout_run_from_pointer(self):
        layout = Layout(self.pango_context)
        layout.set_text('Hello World')
        layout_iter = layout.get_iter()

        layout_run = layout_iter.get_run()
        same_layout = LayoutRun.from_pointer(layout_run.get_pointer())

        assert isinstance(same_layout, LayoutRun)

    def test_layout_run_from_null_pointer(self):
        with self.assertRaises(ValueError):
            LayoutRun.from_pointer(ffi.NULL)

    def test_layout_run_properties(self):
        layout = Layout(self.pango_context)
        layout.set_text('Hello World')
        layout_iter = layout.get_iter()

        layout_run = layout_iter.get_run()
        layout_run_item = layout_run.item
        assert isinstance(layout_run_item, Item)



from . import pango


def units_to_double(i: int) -> float:
    """
    Converts a number in Pango units to floating-point: divides it by
    ``PANGO_SCALE``.

    :param i:
        value in Pango units
    """
    return pango.pango_units_to_double(i)


def units_from_double(d: float) -> int:
    """
    Converts a floating-point number to Pango units: multiplies it by
    ``PANGO_SCALE`` and rounds to nearest integer.

    :param d:
        double floating-point value
    """
    return pango.pango_units_from_double(d)



from pangocffi import Layout
from .context_creator import ContextCreator
import unittest


class TestGlyphItemIterWithContext(unittest.TestCase):

    def setUp(self):
        self.pango_context = ContextCreator.get_pango_context()

    def tearDown(self):
        self.pango_context = None
        ContextCreator.free()

    def test_glyph_item_copies(self):
        text = 'Hi from Παν語'
        layout = Layout(self.pango_context)
        layout.set_text(text)
        layout_iter = layout.get_iter()
        layout_run = layout_iter.get_run()

        assert layout_run.copy().get_pointer() != layout_run.get_pointer()

    def test_glyph_item_splits(self):
        text = 'Hi from Παν語'
        layout = Layout(self.pango_context)
        layout.set_text(text)
        layout_iter = layout.get_iter()
        layout_run = layout_iter.get_run()
        layout_run_hi = layout_run.split(text, 3)

        # Substring "Hi "
        assert layout_run_hi.item.num_chars == 3
        assert layout_run_hi.item.offset == 0

        # Substring "from "
        assert layout_run.item.num_chars == 5
        assert layout_run.item.offset == 3

    def test_glyph_item_get_logical_widths(self):
        text = 'Hi from Παν語'
        layout = Layout(self.pango_context)
        layout.set_text(text)
        layout_iter = layout.get_iter()
        layout_run = layout_iter.get_run()
        logical_widths = layout_run.get_logical_widths(text)

        assert len(logical_widths) == 8
        assert isinstance(logical_widths[0], int)



from . import ffi


class Item:
    """
    An :class:`Item` structure stores information about a segment of text.
    """

    def _init_pointer(self, pointer: ffi.CData):
        self._pointer = pointer

    @classmethod
    def from_pointer(cls, pointer: ffi.CData) -> 'Item':
        """
        Instantiates a :class:`Item` from a pointer.

        :return:
            the item.
        """
        if pointer == ffi.NULL:
            raise ValueError('Null pointer')
        self = object.__new__(cls)
        cls._init_pointer(self, pointer)
        return self

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to this item.

        :return:
            a pointer to the item.
        """
        return self._pointer

    @property
    def offset(self) -> int:
        """
        :return:
            byte offset of the start of this item in text.
        :type: int
        """
        return self._pointer.offset

    @property
    def length(self) -> int:
        """
        :return:
            length of this item in bytes.
        :type: int
        """
        return self._pointer.length

    @property
    def num_chars(self) -> int:
        """
        :return:
            number of Unicode characters in the item.
        :type: int
        """
        return self._pointer.num_chars



from pangocffi import Context, FontDescription, Layout, Alignment, ffi,\
    EllipsizeMode
import unittest


class TestLayout(unittest.TestCase):
    @staticmethod
    def test_layout_returns_identical_context():
        context = Context()
        layout = Layout(context)
        identical_context = layout.get_context()
        assert identical_context.get_pointer() == context.get_pointer()

    @staticmethod
    def test_layout_get_pointer_returns_identical_layout():
        context = Context()
        layout = Layout(context)
        identical_layout = Layout.from_pointer(layout.get_pointer())
        assert identical_layout.get_pointer() == layout.get_pointer()

    def test_layout_not_implemented_equality(self):
        context = Context()
        layout = Layout(context)
        assert ('not an object' != layout)

    def test_layout_returns_null_from_null_pointer(self):
        with self.assertRaises(ValueError):
            Layout.from_pointer(ffi.NULL)

    @staticmethod
    def test_layout_setting_font_description():
        context = Context()
        layout = Layout(context)

        # Assert that the font description is not set
        assert layout.get_font_description() is None

        # Creating the font description
        desc = FontDescription()
        desc.set_family('sans-serif')
        layout.set_font_description(desc)

        # Verifying the font description was set
        same_desc = layout.get_font_description()
        assert same_desc.get_family() == desc.get_family()

        # Changing the font description
        desc.set_family('serif')
        assert same_desc.get_family() != desc.get_family()

        # Resetting the font description
        layout.set_font_description(None)
        assert layout.get_font_description() is None

    @staticmethod
    def test_layout_setting_properties():
        context = Context()
        layout = Layout(context)

        layout.set_width(300)
        assert layout.get_width() == 300

        layout.set_height(400)
        assert layout.get_height() == 400

        assert layout.get_spacing() == 0
        layout.set_spacing(30)
        assert layout.get_spacing() == 30

        layout.set_alignment(Alignment.CENTER)
        assert layout.get_alignment() is Alignment.CENTER

        layout.set_ellipsize(EllipsizeMode.MIDDLE)
        assert layout.get_ellipsize() is EllipsizeMode.MIDDLE

        ink_rect, logical_rect = layout.get_extents()
        assert logical_rect.width == 0
        assert logical_rect.height == 0

        width, height = layout.get_size()
        assert width == 0
        assert height == 0

        baseline = layout.get_baseline()
        assert baseline == 0

        line_count = layout.get_line_count()
        assert line_count == 1

    @staticmethod
    def test_layout_setting_text():
        context = Context()
        layout = Layout(context)

        layout.set_width(300)

        layout.set_text('Hi from Pango')
        layout.set_markup('<span font="italic 30">Hi from Παν語</span>')



from pangocffi import Context, FontDescription, Gravity, GravityHint, ffi
import unittest


class TestContext(unittest.TestCase):

    def test_context_init_identical_context(self):
        context = Context()
        identical_context = Context.from_pointer(context.get_pointer())
        assert identical_context == context

    def test_layout_not_implemented_equality(self):
        context = Context()
        assert ('not an object' != context)

    def test_context_returns_null_from_null_pointer(self):
        with self.assertRaises(ValueError):
            Context.from_pointer(ffi.NULL)

    def test_context_properties(self):
        context = Context()

        desc = FontDescription()
        desc.set_family('sans-serif')
        context.set_font_description(desc)
        assert context.get_font_description().get_family() == 'sans-serif'

        context.set_base_gravity(Gravity.EAST)
        assert context.get_base_gravity() == Gravity.EAST
        assert context.get_gravity() == Gravity.EAST

        context.set_gravity_hint(GravityHint.STRONG)
        assert context.get_gravity_hint() == GravityHint.STRONG



from pangocffi import Layout
from .context_creator import ContextCreator
import unittest


class TestLayoutWithContext(unittest.TestCase):
    def setUp(self):
        self.pango_context = ContextCreator.get_pango_context()

    def tearDown(self):
        self.pango_context = None
        ContextCreator.free()

    def test_layout_setting_text(self):
        layout = Layout(self.pango_context)

        layout.set_text('Hi from Παν語')
        assert layout.get_text() == 'Hi from Παν語'

        layout.set_markup('<span font="italic 30">Hello from Παν語</span>')
        assert layout.get_text() == 'Hello from Παν語'



from pangocffi import Item, ffi
import unittest


class TestItem(unittest.TestCase):

    def test_item_returns_null_from_null_pointer(self):
        with self.assertRaises(ValueError):
            Item.from_pointer(ffi.NULL)



from pangocffi import pango
from enum import Enum


class Style(Enum):
    """
    An enumeration specifying the various slant styles possible for a font.
    """
    NORMAL = pango.PANGO_STYLE_NORMAL
    """the font is upright."""
    OBLIQUE = pango.PANGO_STYLE_OBLIQUE
    """the font is slanted, but in a roman style."""
    ITALIC = pango.PANGO_STYLE_ITALIC
    """the font is slanted in an italic style."""


class Weight(Enum):
    """
    An enumeration specifying the weight (boldness) of a font. This is a
    numerical value ranging from 100 to 1000, but there are some predefined
    values.
    """
    THIN = pango.PANGO_WEIGHT_THIN
    """the thin weight( = 100; Since: 1.24)"""
    ULTRALIGHT = pango.PANGO_WEIGHT_ULTRALIGHT
    """the ultralight weight( = 200)"""
    LIGHT = pango.PANGO_WEIGHT_LIGHT
    """the light weight( = 300)"""
    SEMILIGHT = pango.PANGO_WEIGHT_SEMILIGHT
    """the semilight weight( = 350; Since: 1.36.7)"""
    BOOK = pango.PANGO_WEIGHT_BOOK
    """the book weight( = 380; Since: 1.24)"""
    NORMAL = pango.PANGO_WEIGHT_NORMAL
    """the default weight( = 400)"""
    MEDIUM = pango.PANGO_WEIGHT_MEDIUM
    """the normal weight( = 500; Since: 1.24)"""
    SEMIBOLD = pango.PANGO_WEIGHT_SEMIBOLD
    """the semibold weight( = 600)"""
    BOLD = pango.PANGO_WEIGHT_BOLD
    """the bold weight( = 700)"""
    ULTRABOLD = pango.PANGO_WEIGHT_ULTRABOLD
    """the ultrabold weight( = 800)"""
    HEAVY = pango.PANGO_WEIGHT_HEAVY
    """the heavy weight( = 900)"""
    ULTRAHEAVY = pango.PANGO_WEIGHT_ULTRAHEAVY
    """the ultraheavy weight( = 1000; Since: 1.24)"""


class Variant(Enum):
    """
    An enumeration specifying capitalization variant of the font.
    """
    NORMAL = pango.PANGO_VARIANT_NORMAL
    """A normal font."""
    SMALL_CAPS = pango.PANGO_VARIANT_SMALL_CAPS
    """
    A font with the lower case characters replaced by smaller variants of
    the capital characters.
    """


class Stretch(Enum):
    """
    An enumeration specifying the width of the font relative to other designs
    within a family.
    """
    ULTRA_CONDENSED = pango.PANGO_STRETCH_ULTRA_CONDENSED
    """Ultra Condensed width"""
    EXTRA_CONDENSED = pango.PANGO_STRETCH_EXTRA_CONDENSED
    """Extra Condensed width"""
    CONDENSED = pango.PANGO_STRETCH_CONDENSED
    """Condensed width"""
    SEMI_CONDENSED = pango.PANGO_STRETCH_SEMI_CONDENSED
    """Semi condensed width"""
    NORMAL = pango.PANGO_STRETCH_NORMAL
    """The normal width"""
    SEMI_EXPANDED = pango.PANGO_STRETCH_SEMI_EXPANDED
    """Semi expanded width"""
    EXPANDED = pango.PANGO_STRETCH_EXPANDED
    """Expanded width"""
    EXTRA_EXPANDED = pango.PANGO_STRETCH_EXTRA_EXPANDED
    """Extra Expanded width"""
    ULTRA_EXPANDED = pango.PANGO_STRETCH_ULTRA_EXPANDED
    """Ultra Expanded width"""


class FontMask(Enum):
    """
    The bits in a :class:`FontMask` correspond to fields in a
    :class:`FontDescription` that have been set.
    """
    FAMILY = pango.PANGO_FONT_MASK_FAMILY
    """the font family is specified."""
    STYLE = pango.PANGO_FONT_MASK_STYLE
    """the font style is specified."""
    VARIANT = pango.PANGO_FONT_MASK_VARIANT
    """the font variant is specified."""
    WEIGHT = pango.PANGO_FONT_MASK_WEIGHT
    """the font weight is specified."""
    STRETCH = pango.PANGO_FONT_MASK_STRETCH
    """the font stretch is specified."""
    SIZE = pango.PANGO_FONT_MASK_SIZE
    """the font size is specified."""
    GRAVITY = pango.PANGO_FONT_MASK_GRAVITY
    """the font gravity is specified(Since: 1.16.)"""


class Alignment(Enum):
    """
    :class:`Alignment` describes how to align the lines of a :class:`Layout`
    within the available space. If the :class:`Layout` is set to justify using
    :meth:`set_justify()`, this only has effect for partial lines.
    """
    LEFT = pango.PANGO_ALIGN_LEFT
    """Put all available space on the right"""
    CENTER = pango.PANGO_ALIGN_CENTER
    """Center the line within the available space"""
    RIGHT = pango.PANGO_ALIGN_RIGHT
    """Put all available space on the left"""


class EllipsizeMode(Enum):
    """
    :class:`EllipsizeMode` describes what sort of (if any) ellipsization
    should be applied to a line of text. In the ellipsization process
    characters are removed from the text in order to make it fit to a given
    width and replaced with an ellipsis.
    """
    NONE = pango.PANGO_ELLIPSIZE_NONE
    """No ellipsization"""
    START = pango.PANGO_ELLIPSIZE_START
    """Omit characters at the start of the text"""
    MIDDLE = pango.PANGO_ELLIPSIZE_MIDDLE
    """Omit characters in the middle of the text"""
    END = pango.PANGO_ELLIPSIZE_END
    """Omit characters at the end of the text"""


class Gravity(Enum):
    """
    :class:`Gravity` represents the orientation of glyphs in a segment of text.
    This is useful when rendering vertical text layouts. In those situations,
    the layout is rotated using a non-identity :class:`Matrix`, and then glyph
    orientation is controlled using :class:`Gravity`. Not every value in this
    enumeration makes sense for every usage of :class:`Gravity`; for example,
    ``Gravity.AUTO`` only can be passed to ``Context.set_base_gravity()`` and
    can only be returned by ``Context.get_base_gravity()``.
    """
    SOUTH = pango.PANGO_GRAVITY_SOUTH
    """Glyphs stand upright (default)"""
    EAST = pango.PANGO_GRAVITY_EAST
    """Glyphs are rotated 90 degrees clockwise"""
    NORTH = pango.PANGO_GRAVITY_NORTH
    """Glyphs are upside-down"""
    WEST = pango.PANGO_GRAVITY_WEST
    """Glyphs are rotated 90 degrees counter-clockwise"""
    AUTO = pango.PANGO_GRAVITY_AUTO
    """Gravity is resolved from the context matrix"""


class GravityHint(Enum):
    """
    :class:`GravityHint` defines how horizontal scripts should behave in a
    vertical context. That is, English excerpt in a vertical paragraph for
    example.

    See :class:`Gravity`.
    """
    NATURAL = pango.PANGO_GRAVITY_HINT_NATURAL
    """
    scripts will take their natural gravity based on the base gravity and the
    script. This is the default.
    """
    STRONG = pango.PANGO_GRAVITY_HINT_STRONG
    """always use the base gravity set, regardless of the script."""
    LINE = pango.PANGO_GRAVITY_HINT_LINE
    """
    for scripts not in their natural direction (eg. Latin in East gravity),
    choose per-script gravity such that every script respects the line
    progression. This means, Latin and Arabic will take opposite gravities and
    both flow top-to-bottom for example.
    """



from pangocffi import Layout, GlyphItemIter, GlyphItem, ffi
from .context_creator import ContextCreator
import unittest


class TestGlyphItemIter(unittest.TestCase):

    def setUp(self):
        self.pango_context = ContextCreator.get_pango_context()

    def tearDown(self):
        self.pango_context = None
        ContextCreator.free()

    def test_glyph_item_iterator(self):
        text = 'Hi from Παν語'
        layout = Layout(self.pango_context)
        layout.set_text(text)
        layout_iter = layout.get_iter()

        glyph_item_iter = GlyphItemIter()
        assert isinstance(glyph_item_iter.get_pointer(), ffi.CData)

        assert glyph_item_iter.init_start(
            layout_iter.get_run(),
            text
        )

        assert isinstance(glyph_item_iter.glyph_item, GlyphItem)
        assert glyph_item_iter.text == text

        # Verify first cluster
        assert glyph_item_iter.start_glyph == 0
        assert glyph_item_iter.start_index == 0
        assert glyph_item_iter.start_char == 0
        assert glyph_item_iter.end_glyph == 1
        assert glyph_item_iter.end_index == 1
        assert glyph_item_iter.end_char == 1

        assert not glyph_item_iter.prev_cluster()
        assert glyph_item_iter.next_cluster()

        # Verify second cluster
        assert glyph_item_iter.start_glyph == 1
        assert glyph_item_iter.start_index == 1
        assert glyph_item_iter.start_char == 1
        assert glyph_item_iter.end_glyph == 2
        assert glyph_item_iter.end_index == 2
        assert glyph_item_iter.end_char == 2

        assert glyph_item_iter.init_end(
            layout_iter.get_run(),
            text
        )

        # Verify last cluster
        assert glyph_item_iter.start_glyph == 7
        assert glyph_item_iter.start_index == 7
        assert glyph_item_iter.start_char == 7
        assert glyph_item_iter.end_glyph == 8
        assert glyph_item_iter.end_index == 8
        assert glyph_item_iter.end_char == 8

        assert not glyph_item_iter.next_cluster()
        assert glyph_item_iter.prev_cluster()

        # Verify second to last cluster
        assert glyph_item_iter.start_glyph == 6
        assert glyph_item_iter.start_index == 6
        assert glyph_item_iter.start_char == 6
        assert glyph_item_iter.end_glyph == 7
        assert glyph_item_iter.end_index == 7
        assert glyph_item_iter.end_char == 7



from . import pango, gobject, ffi, FontDescription, Gravity, GravityHint


class Context(object):
    """
    The :class:`Context` structure stores global information used to control
    the itemization process.
    """

    def __init__(self):
        """
        Creates a new PangoContext initialized to default values.

        This function is not particularly useful as it should always be
        followed by a :meth:`set_font_map()` call, and the function
        ``FontMap.create_context()`` does these two steps together and hence
        users are recommended to use that.

        If you are using Pango as part of a higher-level system, that system
        may have it's own way of create a :class:`Context`. For instance, the
        PangoCairo toolkit has pango_cairo_create_context(). Use those instead.
        """
        self._init_pointer(pango.pango_context_new())

    def _init_pointer(self, pointer: ffi.CData):
        self._pointer = ffi.gc(pointer, gobject.g_object_unref)

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to the context

        :return:
            the pointer to the context.
        """
        return self._pointer

    @classmethod
    def from_pointer(cls, pointer: ffi.CData) -> 'Context':
        """
        Instantiates a :class:`Context` from a pointer.

        :return:
            the context.
        """
        if pointer == ffi.NULL:
            raise ValueError('Null pointer')
        self = object.__new__(cls)
        cls._init_pointer(self, pointer)
        return self

    def __eq__(self, other):
        if isinstance(other, Context):
            return self.get_pointer() == other.get_pointer()
        return NotImplemented

    def set_font_description(self, desc: FontDescription) -> None:
        """
        Set the default font description for the context.

        :param desc:
            the new pango font description.
        """
        pango.pango_context_set_font_description(
            self._pointer, desc.get_pointer()
        )

    def get_font_description(self) -> FontDescription:
        """
        Retrieve the default font description for the context.

        :return:
            the context's default font description.
        """
        return FontDescription.from_pointer(
            pango.pango_context_get_font_description(self._pointer)
        )

    def get_base_gravity(self) -> Gravity:
        """
        Retrieves the base gravity for the context.
        See :meth:`set_base_gravity()`.

        :return:
            the base gravity for the context.
        """
        return Gravity(
            pango.pango_context_get_base_gravity(self._pointer)
        )

    def set_base_gravity(self, gravity: Gravity):
        """
        Sets the base gravity for the context.

        The base gravity is used in laying vertical text out.

        :param gravity:
            the new base gravity
        """
        pango.pango_context_set_base_gravity(self._pointer, gravity.value)

    def get_gravity(self) -> Gravity:
        """
        Retrieves the gravity for the context. This is similar to
        :meth:`get_base_gravity()`, except for when the base gravity is
        :meth:`Gravity.AUTO` for which :meth:`get_for_matrix()` is used to
        return the gravity from the current context matrix.

        :return:
            the resolved gravity for the context.
        """
        return Gravity(
            pango.pango_context_get_gravity(self._pointer)
        )

    def get_gravity_hint(self) -> GravityHint:
        """
        Retrieves the gravity hint for the context.
        See :meth:`set_gravity_hint()` for details.

        :return:
            the gravity hint for the context.
        """
        return GravityHint(
            pango.pango_context_get_base_gravity(self._pointer)
        )

    def set_gravity_hint(self, hint: GravityHint):
        """
        Sets the gravity hint for the context.

        The gravity hint is used in laying vertical text out, and is only
        relevant if gravity of the context as returned by
        :meth:`get_gravity()` is set :meth:`Gravity.EAST` or
        :meth:`Gravity.WEST`.

        :param hint:
            the new gravity hint
        """
        pango.pango_context_set_gravity_hint(self._pointer, hint.value)



from . import pango, ffi
from . import GlyphItem


class GlyphItemIter:
    """
    A :class:`GlyphItemIter` is an iterator over the clusters in a
    :class:`GlyphItem`. The forward direction of the iterator is the logical
    direction of text. That is, with increasing ``start_index`` and
    ``start_char`` values. If ``glyph_item`` is right-to-left (that is, if
    ``glyph_item->item->analysis.level`` is odd), then ``start_glyph``
    decreases as the iterator moves forward. Moreover, in right-to-left cases,
    ``start_glyph`` is greater than ``end_glyph``.

    Note that ``text`` is the start of the text for layout, which is then
    indexed by ``glyph_item->item->offset`` to get to the text of
    ``glyph_item``. The ``start_index`` and ``end_index`` values can directly
    index into ``text``. The ``start_glyph``, ``end_glyph``, ``start_char``,
    and ``end_char`` values however are zero-based for the ``glyph_item``.
    For each cluster, the item pointed at by the start variables is included in
    the cluster while the one pointed at by end variables is not.

    None of the members of a :class:`GlyphItemIter` should be modified
    manually.
    """

    def __init__(self):
        """
        Returns an iterator to iterate over the visual extents of the layout.

        :param layout:
            a Pango :class:`Layout`
        """
        self._pointer = ffi.new("PangoGlyphItemIter *")

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to this iterator.

        :return:
            a pointer to the iterator.
        """
        return self._pointer

    def init_start(self, glyph_item: GlyphItem, text: str) -> bool:
        """
        Initializes the :class:`GlyphItemIter` structure to point to the first
        cluster in a glyph item. See :class:`GlyphItemIter` for details of
        cluster orders.

        :return:
            ``False`` if there are no clusters in the glyph item
        """
        text_pointer = ffi.new('char[]', text.encode('utf8'))
        return bool(pango.pango_glyph_item_iter_init_start(
            self._pointer,
            glyph_item.get_pointer(),
            text_pointer
        ))

    def init_end(self, glyph_item: GlyphItem, text: str) -> bool:
        """
        Initializes the :class:`GlyphItemIter` structure to point to the last
        cluster in a glyph item. See :class:`GlyphItemIter` for details of
        cluster orders.

        :return:
            ``False`` if there are no clusters in the glyph item
        """
        text_pointer = ffi.new('char[]', text.encode('utf8'))
        return bool(pango.pango_glyph_item_iter_init_end(
            self._pointer,
            glyph_item.get_pointer(),
            text_pointer
        ))

    def next_cluster(self) -> bool:
        """
        Moves the iterator forward to the next run in visual order. If the
        iterator was already at the end of the layout, it will
        return ``False``.

        :return:
            ``True`` if the iterator was advanced, ``False`` if we were
            already on the last cluster.
        """
        return bool(pango.pango_glyph_item_iter_next_cluster(self._pointer))

    def prev_cluster(self) -> bool:
        """
        Moves the iterator to the preceding cluster in the glyph item. See
        :class:`GlyphItemIter` for details of cluster orders.

        :return:
            ``True`` if the iterator was moved, ``False`` if we were already
            on the first cluster.
        """
        return bool(pango.pango_glyph_item_iter_prev_cluster(self._pointer))

    @property
    def glyph_item(self) -> GlyphItem:
        """
        :return:
            the glyph item being iterated over
        :type: GlyphItem
        """
        return GlyphItem.from_pointer(self._pointer.glyph_item)

    @property
    def text(self) -> str:
        """
        :return:
            the text for the layout
        :type: str
        """
        return ffi.string(self._pointer.text).decode('utf-8')

    @property
    def start_glyph(self) -> int:
        """
        :type: int
        """
        return self._pointer.start_glyph

    @property
    def start_index(self) -> int:
        """
        :return:
            the cluster start index within text
        :type: int
        """
        return self._pointer.start_index

    @property
    def start_char(self) -> int:
        """
        :type: int
        """
        return self._pointer.start_char

    @property
    def end_glyph(self) -> int:
        """
        :type: int
        """
        return self._pointer.end_glyph

    @property
    def end_index(self) -> int:
        """
        :return:
            the cluster end index within text
        :type: int
        """
        return self._pointer.end_index

    @property
    def end_char(self) -> int:
        """
        :type: int
        """
        return self._pointer.end_char



from typing import Tuple, Optional
from . import pango, ffi
from . import LayoutRun, Rectangle


class LayoutIter:
    """
    A :class:`LayoutIter` can be used to iterate over the visual extents of a
    Pango :class:`Layout`.
    """

    def _init_pointer(self, pointer: ffi.CData):
        self._pointer = ffi.gc(pointer, pango.pango_layout_iter_free)

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to this iterator.

        :return:
            a pointer to the iterator.
        """
        return self._pointer

    @classmethod
    def from_pointer(cls, pointer: ffi.CData) -> 'LayoutIter':
        """
        Instantiates a :class:`LayoutIter` from a pointer.

        :return:
            the layout iterator.
        """
        if pointer == ffi.NULL:
            raise ValueError('Null pointer')
        self = object.__new__(cls)
        cls._init_pointer(self, pointer)
        return self

    def next_run(self) -> bool:
        """
        Moves the iterator forward to the next run in visual order. If the
        iterator was already at the end of the layout, it will
        return ``False``.

        :return:
            whether motion was possible.
        """
        return bool(pango.pango_layout_iter_next_run(self._pointer))

    def next_char(self) -> bool:
        """
        Moves the iterator forward to the next character in visual order. If
        the iterator was already at the end of the layout, it will
        return ``False``.

        :return:
            whether motion was possible.
        """
        return bool(pango.pango_layout_iter_next_char(self._pointer))

    def next_cluster(self) -> bool:
        """
        Moves the iterator forward to the next cluster in visual order. If
        the iterator was already at the end of the layout, it will
        return ``False``.

        :return:
            whether motion was possible.
        """
        return bool(pango.pango_layout_iter_next_cluster(self._pointer))

    def next_line(self) -> bool:
        """
        Moves the iterator forward to the start of the next line. If
        the iterator was already at the end of the layout, it will
        return ``False``.

        :return:
            whether motion was possible.
        """
        return bool(pango.pango_layout_iter_next_line(self._pointer))

    def at_last_line(self) -> bool:
        """
        Determines whether the iterator is on the last line of the layout.

        :return:
            ``True`` if iterator is on the last line.
        """
        return bool(pango.pango_layout_iter_at_last_line(self._pointer))

    def get_index(self) -> int:
        """
        Returns the current byte index. Note that iterating forward by char
        moves in visual order, not logical order, so indexes may not be
        sequential. Also, the index may be equal to the length of the text
        in the layout, if on the NULL run (see :meth:`get_run()`).

        :return:
            The current byte index.
        """
        return pango.pango_layout_iter_get_index(self._pointer)

    def get_baseline(self) -> int:
        """
        Returns the Y position of the current line's baseline, in layout
        coordinates (origin at top left of the entire layout).

        :return:
            The baseline of the current line.
        """
        return pango.pango_layout_iter_get_baseline(self._pointer)

    def get_run(self) -> Optional[LayoutRun]:
        """
        Returns the current run. When iterating by run, at the end of each
        line, there's a position with a NULL run, so this function can return
        None. The NULL run at the end of each line ensures that all lines have
        at least one run, even lines consisting of only a newline.

        Use the faster :meth:`get_run_readonly()` if you do not plan
        to modify the contents of the run (glyphs, glyph widths, etc.).

        :return:
            the current run
        """
        run_pointer = pango.pango_layout_iter_get_run(self._pointer)
        if run_pointer == ffi.NULL:
            return None
        return LayoutRun.from_pointer(run_pointer)

    # def get_line(_readonly)

    # def get_layout

    def get_char_extents(self) -> Rectangle:
        """
        Returns the extents of the current character, in layout coordinates
        (origin is the top left of the entire layout). Only logical extents
        can sensibly be obtained for characters; ink extents make sense only
        down to the level of clusters.

        :return:
             a rectangle representing the logical extent of a character.
        """
        logical_rect = Rectangle()
        pango.pango_layout_iter_get_char_extents(
            self._pointer,
            logical_rect.get_pointer()
        )
        return logical_rect

    def get_cluster_extents(self) -> Tuple[Rectangle, Rectangle]:
        """
        Returns the extents of the current cluster, in layout coordinates
        (origin is the top left of the entire layout).

        :return:
            a tuple containing two :class:`Rectangle` objects.
            The first is the extent of the cluster as drawn.
            The second is the logical extent.
        """
        ink_rect = Rectangle()
        logical_rect = Rectangle()
        pango.pango_layout_iter_get_cluster_extents(
            self._pointer,
            ink_rect.get_pointer(),
            logical_rect.get_pointer()
        )
        return ink_rect, logical_rect

    def get_run_extents(self) -> Tuple[Rectangle, Rectangle]:
        """
        Returns the extents of the current run, in layout coordinates (origin
        is the top left of the entire layout).

        :return:
            a tuple containing two :class:`Rectangle` objects.
            The first is the extent of the run as drawn.
            The second is the logical extent.
        """
        ink_rect = Rectangle()
        logical_rect = Rectangle()
        pango.pango_layout_iter_get_run_extents(
            self._pointer,
            ink_rect.get_pointer(),
            logical_rect.get_pointer()
        )
        return ink_rect, logical_rect

    def get_line_yrange(self) -> Tuple[int, int]:
        """
        Divides the vertical space in the :class:`Layout` being iterated over
        between the lines in the layout, and returns the space belonging to
        the current line. A line's range includes the line's logical
        extents, plus half of the spacing above and below the line, if
        :meth:`Layout.set_spacing()` has been called to set layout spacing.
        The Y positions are in layout coordinates (origin at top left of
        the entire layout).

        :return:
            a tuple containing two integers.
            The first is the y position of the start of the line.
            The second is the y position of the end of the line
        """
        y0_pointer = ffi.new("int *")
        y1_pointer = ffi.new("int *")
        pango.pango_layout_iter_get_line_yrange(
            self._pointer,
            y0_pointer,
            y1_pointer
        )
        y0 = y0_pointer[0]
        y1 = y1_pointer[0]
        return y0, y1

    def get_line_extents(self) -> Tuple[Rectangle, Rectangle]:
        """
        Obtains the extents of the current line. Extents are in layout
        coordinates (origin is the top-left corner of the entire
        :class:`Layout`). Thus the extents returned by this function will be
        the same width/height but not at the same x/y as the extents returned
        from :meth:`LayoutLine.get_extents()`.

        :return:
            a tuple containing two :class:`Rectangle` objects.
            The first is the extent of the line as drawn.
            The second is the logical extent.
        """
        ink_rect = Rectangle()
        logical_rect = Rectangle()
        pango.pango_layout_iter_get_line_extents(
            self._pointer,
            ink_rect.get_pointer(),
            logical_rect.get_pointer()
        )
        return ink_rect, logical_rect

    def get_layout_extents(self) -> Tuple[Rectangle, Rectangle]:
        """
        Returns the extents of the :class:`Layout` being iterated over.

        :return:
            a tuple containing two :class:`Rectangle` objects.
            The first is the extent of the run as drawn.
            The second is the logical extent.
        """
        ink_rect = Rectangle()
        logical_rect = Rectangle()
        pango.pango_layout_iter_get_layout_extents(
            self._pointer,
            ink_rect.get_pointer(),
            logical_rect.get_pointer()
        )
        return ink_rect, logical_rect



from typing import List
from . import pango, ffi, Item


class GlyphItem:
    """
    A :class:`GlyphItem` is a pair of a :class:`Item` and the glyphs resulting
    from shaping the text corresponding to an item. As an example of the usage
    of :class:`GlyphItem`, the results of shaping text with :class:`Layout` is
    a list of :class:`LayoutLine`, each of which contains a list of
    :class:`GlyphItem`.
    """

    def _init_pointer(self, pointer: ffi.CData):
        self._pointer = pointer

    @classmethod
    def from_pointer(cls, pointer: ffi.CData) -> 'GlyphItem':
        """
        Instantiates a :class:`GlyphItem` from a pointer.

        :return:
            the glyph item.
        """
        if pointer == ffi.NULL:
            raise ValueError('Null pointer')
        self = object.__new__(cls)
        cls._init_pointer(self, pointer)
        return self

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to this glyph item.

        :return:
            a pointer to the glyph item.
        """
        return self._pointer

    def copy(self) -> 'GlyphItem':
        """
        Make a deep copy of the :class:`GlyphItem` structure.

        :return:
            a copy of the :class:`GlyphItem`
        """
        glyph_item_pointer = pango.pango_glyph_item_copy(self._pointer)
        glyph_item_pointer = ffi.gc(
            glyph_item_pointer,
            pango.pango_glyph_item_free
        )
        return GlyphItem.from_pointer(glyph_item_pointer)

    def split(self, text: str, split_index: int) -> 'GlyphItem':
        """
        Modifies ``orig`` to cover only the text after ``split_index``, and
        returns a new item that covers the text before ``split_index`` that
        used to be in ``orig``. You can think of ``split_index`` as the length
        of the returned item. ``split_index`` may not be 0, and it may not be
        greater than or equal to the length of orig (that is, there must be at
        least one byte assigned to each item, you can't create a zero-length
        item).

        This function is similar in function to :meth:`Item.split()` (and uses
        it internally.)

        :param text:
            text to which positions in ``orig`` apply
        :param split_index:
            byte index of position to split item, relative to the start of the
            item
        :return:
             a new :class:`GlyphItem` representing text before ``split_index``.
        """
        text_pointer = ffi.new('char[]', text.encode('utf8'))
        glyph_item_pointer = pango.pango_glyph_item_split(
            self._pointer,
            text_pointer,
            split_index
        )
        glyph_item_pointer = ffi.gc(
            glyph_item_pointer,
            pango.pango_glyph_item_free
        )
        return GlyphItem.from_pointer(glyph_item_pointer)

    # Todo: apply_attrs(self, text: str, list: AttrList)

    # Todo:
    # def letter_space(
    #    self,
    #    text: str,
    #    log_attrs: LogAttr,
    #    letter_spacing: int
    # )

    def get_logical_widths(self, text: str) -> List[int]:
        """
        Given a :class:`GlyphItem` and the corresponding text, determine the
        screen width corresponding to each character. When multiple characters
        compose a single cluster, the width of the entire cluster is divided
        equally among the characters.

        See also :meth:`GlyphString.get_logical_widths()`.

        :param text:
            text that ``glyph_item`` corresponds to (
            ``glyph_item->item->offset`` is an offset from the start of
            ``text``)
        :return:
            an array whose length is the number of characters in `glyph_item`
            (equal to ``glyph_item->item->num_chars``) to be filled in with
            the resulting character widths.)
        """
        logical_widths_pointer = ffi.new('int [%d]' % self.item.num_chars)
        text_pointer = ffi.new('char[]', text.encode('utf8'))
        pango.pango_glyph_item_get_logical_widths(
            self._pointer,
            text_pointer,
            logical_widths_pointer
        )

        # type = ffi.typeof(logical_widths_pointer)
        return [logical_widths_pointer[i] for i in range(self.item.num_chars)]

    @property
    def item(self) -> Item:
        """
        :return:
            corresponding :class:`Item`.
        :type: Item
        """
        return Item.from_pointer(self._pointer.item)



from pangocffi import Rectangle


def test_rectangle_properties():
    rect = Rectangle()
    assert rect.x == 0
    assert rect.y == 0
    assert rect.width == 0
    assert rect.height == 0


def test_rectangle_pointers():
    rect = Rectangle()
    same_rect = rect.from_pointer(rect.get_pointer())
    assert same_rect.x == 0



"""
    pangocffi.ffi_instance_builder
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    generates an FFI for pangocffi
"""

from pathlib import Path
from cffi import FFI
from typing import Optional


class FFIInstanceBuilder:

    def __init__(self, source: Optional[str] = None):
        self.source = source

    def generate(self) -> FFI:
        # Read the C definitions
        c_definitions_glib_file = open(
            str(Path(__file__).parent / 'c_definitions_glib.txt'),
            'r'
        )
        c_definitions_pango_file = open(
            str(Path(__file__).parent / 'c_definitions_pango.txt'),
            'r'
        )
        c_definitions_glib = c_definitions_glib_file.read()
        c_definitions_pango = c_definitions_pango_file.read()

        ffi = FFI()
        ffi.cdef(c_definitions_glib)
        ffi.cdef(c_definitions_pango)
        if self.source is not None:
            ffi.set_source(self.source, None)

        return ffi



from pangocffi import FontDescription
from pangocffi import Gravity, Stretch, Style, Variant, Weight, ffi
import unittest


class TestFontDescription(unittest.TestCase):

    def test_font_description_init_identical(self):
        desc = FontDescription()
        identical_desc = desc.from_pointer(desc.get_pointer())
        assert identical_desc == desc

    def test_font_description_not_implemented_equality(self):
        desc = FontDescription()
        assert ('not an object' != desc)

    def test_font_description_returns_null_from_null_pointer(self):
        with self.assertRaises(ValueError):
            FontDescription.from_pointer(ffi.NULL)

    def test_setting_properties(self):
        desc = FontDescription()

        assert desc.get_family() is None
        desc.set_family('sans-serif')
        assert desc.get_family() == 'sans-serif'

        desc.set_style(Style.NORMAL)
        assert desc.get_style() == Style.NORMAL

        desc.set_variant(Variant.NORMAL)
        assert desc.get_variant() == Variant.NORMAL

        desc.set_weight(700)
        assert desc.get_weight() == 700
        desc.set_weight(Weight.BOOK)
        assert desc.get_weight() == Weight.BOOK.value

        desc.set_stretch(Stretch.NORMAL)
        assert desc.get_stretch() == Stretch.NORMAL

        desc.set_size(123)
        assert desc.get_size() == 123
        assert not desc.get_size_is_absolute()

        desc.set_absolute_size(1.23)
        assert desc.get_size() != 123
        assert desc.get_size_is_absolute()

        desc.set_gravity(Gravity.EAST)
        assert desc.get_gravity() == Gravity.EAST



import ctypes.util
from ._generated.ffi import ffi


def _dlopen(generated_ffi, *names):
    """Try various names for the same library, for different platforms."""
    for name in names:
        for lib_name in (name, 'lib' + name):
            try:
                path = ctypes.util.find_library(lib_name)
                lib = generated_ffi.dlopen(path or lib_name)
                if lib:
                    return lib
            except OSError:
                pass
    raise OSError("dlopen() failed to load a library: %s" % ' / '.join(names))


pango = _dlopen(ffi, 'pango', 'pango-1', 'pango-1.0', 'pango-1.0-0')
gobject = _dlopen(ffi, 'gobject-2.0', 'gobject-2.0-0')


# Imports are normally always put at the top of the file.
# But the wrapper API requires that the pango library be loaded first.
# Therefore, we have to disable linting rules for these lines.
from .version import *  # noqa
from .enums import *  # noqa
from .convert import *  # noqa
from .font_description import FontDescription  # noqa
from .rectangle import Rectangle  # noqa
from .item import Item  # noqa
from .context import Context  # noqa
from .glyph_item import GlyphItem  # noqa
from .glyph_item_iter import GlyphItemIter  # noqa
from .layout_run import LayoutRun  # noqa
from .layout_iter import LayoutIter  # noqa
from .layout import Layout  # noqa



from pangocffi import Context, Layout, LayoutIter, Rectangle, ffi
import unittest


class TestLayoutIter(unittest.TestCase):

    @staticmethod
    def test_layout_iter_pointer():
        context = Context()
        layout = Layout(context)
        layout_iter = layout.get_iter()
        assert isinstance(layout_iter.get_pointer(), ffi.CData)

    def test_layout_iter_returns_null_from_null_pointer(self):
        with self.assertRaises(ValueError):
            LayoutIter.from_pointer(ffi.NULL)

    @staticmethod
    def test_layout_iter_properties():
        context = Context()
        layout = Layout(context)
        layout_iter = layout.get_iter()

        assert layout_iter.next_run() is False
        assert layout_iter.next_char() is False
        assert layout_iter.next_cluster() is False
        assert layout_iter.next_line() is False
        assert layout_iter.at_last_line() is True
        assert layout_iter.get_index() == 0
        assert layout_iter.get_baseline() == 0

        assert isinstance(layout_iter.get_char_extents(), Rectangle)

        cluster_ink, cluster_logical = layout_iter.get_cluster_extents()
        assert isinstance(cluster_ink, Rectangle)
        assert isinstance(cluster_logical, Rectangle)

        run_ink, run_logical = layout_iter.get_run_extents()
        assert isinstance(run_ink, Rectangle)
        assert isinstance(run_logical, Rectangle)

        y0, y1 = layout_iter.get_line_yrange()
        assert y0 == 0
        assert y1 == 0

        line_ink, line_logical = layout_iter.get_line_extents()
        assert isinstance(line_ink, Rectangle)
        assert isinstance(line_logical, Rectangle)

        layout_ink, layout_logical = layout_iter.get_layout_extents()
        assert isinstance(layout_ink, Rectangle)
        assert isinstance(layout_logical, Rectangle)

    @staticmethod
    def test_layout_iterator_run_returns_none():
        context = Context()
        layout = Layout(context)
        layout_iter = layout.get_iter()

        assert layout_iter.get_run() is None



from . import pango, gobject, ffi
from . import Context, FontDescription
from . import Alignment, Rectangle, EllipsizeMode
from pangocffi import LayoutIter
from typing import Tuple, Optional


class Layout(object):
    """
    A Pango :class:`Layout` represents an entire paragraph of text. It is
    initialized with a Pango :class:`Context`, UTF-8 string and set of
    attributes for that string. Once that is done, the set of formatted lines
    can be extracted from the object, the layout can be rendered, and
    conversion between logical character positions within the layout's text,
    and the physical position of the resulting glyphs can be made.
    """

    def __init__(self, context: Context):
        """
        Create a new :class:`Layout` object with attributes initialized to
        default values for a particular :class:`Context`.

        :param context:
            the Pango :class:`Context`
        """
        self._init_pointer(pango.pango_layout_new(context.get_pointer()))

    def _init_pointer(self, pointer: ffi.CData):
        self._pointer = ffi.gc(pointer, gobject.g_object_unref)

    @classmethod
    def from_pointer(cls, pointer: ffi.CData) -> 'Layout':
        """
        Instantiates a :class:`Layout` from a pointer.

        :return:
            the layout.
        """
        if pointer == ffi.NULL:
            raise ValueError('Null pointer')
        self = object.__new__(cls)
        cls._init_pointer(self, pointer)
        return self

    def get_pointer(self) -> ffi.CData:
        """
        Returns the pointer to this layout.

        :return:
            a pointer to the layout.
        """
        return self._pointer

    def get_context(self) -> Context:
        """
        Returns the Pango :class:`Context` used for this layout.

        :return:
            the :class:`Context` for the layout.
        """
        return Context.from_pointer(
            pango.pango_layout_get_context(self._pointer)
        )

    def set_text(self, text: str) -> None:
        """
        Sets the text of the layout.

        Note that if you have used ``set_markup()`` or
        ``set_markup_with_accel()`` on the layout before, you may want to call
        ``set_attributes()`` to clear the attributes set on the layout from
        the markup as this function does not clear attributes.

        :param text:
            a valid UTF-8 string
        """
        text_pointer = ffi.new('char[]', text.encode('utf8'))
        pango.pango_layout_set_text(self._pointer, text_pointer, -1)

    def get_text(self) -> str:
        """
        Returns the text in the layout.

        :return:
            the text in the layout.
        """
        text_pointer = pango.pango_layout_get_text(self._pointer)
        return ffi.string(text_pointer).decode("utf-8")

    def set_markup(self, markup: str) -> None:
        """
        Same as set_markup_with_accel(), but the markup text isn't scanned for
        accelerators.

        :param markup:
            marked-up text
        """
        markup_pointer = ffi.new('char[]', markup.encode('utf8'))
        pango.pango_layout_set_markup(self._pointer, markup_pointer, -1)

    def set_font_description(self, desc: Optional[FontDescription]) -> None:
        """
        Sets the default font description for the layout. If no font
        description is set on the layout, the font description from the
        layout's context is used.

        :param desc:
            the new :class:`FontDescription`, or ``None`` to unset the current
            ``FontDescription``.
        """
        if desc is None:
            pango.pango_layout_set_font_description(self._pointer, ffi.NULL)
        else:
            pango.pango_layout_set_font_description(
                self._pointer,
                desc.get_pointer()
            )

    def get_font_description(self) -> Optional[FontDescription]:
        """
        Returns the font description for the layout, if any.

        :return:
            the layout's :class:`FontDescription`, or ``None`` if the
            :class:`FontDescription` from the layout's :class:`Context` is
            inherited.
        """
        desc_pointer = pango.pango_layout_get_font_description(self._pointer)
        if desc_pointer == ffi.NULL:
            return None
        return FontDescription.from_pointer(desc_pointer)

    def set_width(self, width: int) -> None:
        """
        Sets the width to which the lines of the layout should wrap or
        ellipsized. The default value is -1: no width set.

        :param width:
            the desired width in Pango units, or -1 to indicate that no
            wrapping or ellipsization should be performed.
        """
        pango.pango_layout_set_width(self._pointer, width)

    def get_width(self) -> int:
        """
        Returns the width to which the lines of the layout should wrap.

        :return:
             the width in Pango units, or -1 if no width set.
        """
        return pango.pango_layout_get_width(self._pointer)

    def set_height(self, height: int) -> None:
        """
        Sets the ``height`` to which the layout should be ellipsized at.
        There are two different behaviors, based on whether height is positive
        or negative.

        If ``height`` is positive, it will be the maximum height of the layout.
        Only lines would be shown that would fit, and if there is any text
        omitted, an ellipsis added. At least one line is included in each
        paragraph regardless of how small the height value is.
        A value of zero will render exactly one line for the entire layout.

        If height is negative, it will be the (negative of) maximum number of
        lines per paragraph. That is, the total number of lines shown may well
        be more than this value if the layout contains multiple paragraphs of
        text. The default value of -1 means that first line of each paragraph
        is ellipsized. This behavior may be changed in the future to act per
        layout instead of per paragraph. File a bug against pango at
        http://bugzilla.gnome.org/ if your code relies on this behavior.

        Height setting only has effect if a positive width is set on layout and
        ellipsization mode of layout is not ``Ellipsize.NONE``. The behavior
        is undefined if a height other than -1 is set and ellipsization mode
        is set to ``Ellipsize.NONE``, and may change in the future.

        :param height:
            the desired height of the layout in Pango units if positive, or
            desired number of lines if negative.
        """
        pango.pango_layout_set_height(self._pointer, height)

    def get_height(self) -> int:
        """
        Returns the height of the layout used for ellipsization.

        :return:
            the height, in Pango units if positive, or number of lines if
            negative.
        """
        return pango.pango_layout_get_height(self._pointer)

    def get_spacing(self) -> int:
        """
        Returns the amount of spacing between the lines of the layout.

        :return:
            the spacing in pango units.
        """
        return pango.pango_layout_get_spacing(self._pointer)

    def set_spacing(self, spacing: int) -> None:
        """
        Sets the amount of spacing in Pango unit between the lines of
        the layout.

        :param spacing:
            the amount of spacing
        """
        pango.pango_layout_set_spacing(self._pointer, spacing)

    def set_alignment(self, alignment: Alignment) -> None:
        """
        Sets the alignment for the layout: how partial lines are positioned
        within the horizontal space available.

        :param alignment:
            the alignment.
        """
        pango.pango_layout_set_alignment(self._pointer, alignment.value)

    def get_alignment(self) -> Alignment:
        """
        Returns the alignment for the layout: how partial lines are positioned
        within the horizontal space available.

        :return:
            the alignment.
        """
        return Alignment(pango.pango_layout_get_alignment(self._pointer))

    def set_ellipsize(self, ellipsize: EllipsizeMode) -> None:
        """
        Sets the ellipsize mode for the layout.

        :param ellipsize:
            the ellipsize mode.
        """

        pango.pango_layout_set_ellipsize(self._pointer, ellipsize.value)

    def get_ellipsize(self) -> EllipsizeMode:
        """
        Returns the ellipsize mode of the layout.

        :return:
            the the ellipsize mode.
        """
        return EllipsizeMode(pango.pango_layout_get_ellipsize(self._pointer))

    def get_extents(self) -> Tuple[Rectangle, Rectangle]:
        """
        Computes the logical and ink extents of the layout.
        Logical extents are usually what you want for positioning things.
        Note that both extents may have non-zero x and y.
        You may want to use those to offset where you render the layout.
        Not doing that is a very typical bug that shows up as right-to-left
        layouts not being correctly positioned in a layout with a set width.

        The extents are given in layout coordinates and in Pango units;
        layout coordinates begin at the top left corner of the layout.

        :return:
            a tuple containing two :class:`Rectangle` objects.
            The first is the extent of the layout as drawn.
            The second is the logical extent of the layout.
        """
        ink_rect = Rectangle()
        logical_rect = Rectangle()
        pango.pango_layout_get_extents(
            self._pointer,
            ink_rect.get_pointer(),
            logical_rect.get_pointer()
        )
        return ink_rect, logical_rect

    def get_size(self) -> Tuple[int, int]:
        """
        Determines the logical width and height of the layout in Pango units.
        This is simply a convenience function around :meth:`get_extents()`.

        :return:
            a tuple containing the logical width and height, respectively.
        """
        width_pointer = ffi.new("int *")
        height_pointer = ffi.new("int *")
        pango.pango_layout_get_size(
            self._pointer,
            width_pointer,
            height_pointer
        )
        width = width_pointer[0]
        height = height_pointer[0]
        return width, height

    def get_baseline(self) -> int:
        """
        Returns the Y position of baseline of the first line in the layout.

        :return:
            baseline of first line, from top of :class:`Layout`
        """
        return pango.pango_layout_get_baseline(self._pointer)

    def get_line_count(self) -> int:
        """
        Returns the number of lines.

        :return:
            the line count
        """
        return pango.pango_layout_get_line_count(self._pointer)

    def get_iter(self) -> LayoutIter:
        """
        Returns an iterator to iterate over the visual extents of the layout.

        :return:
            the layout iterator
        """
        layout_iterator_pointer = pango.pango_layout_get_iter(self._pointer)
        return LayoutIter.from_pointer(layout_iterator_pointer)



import sys
import setuptools

if sys.version_info.major < 3:
    raise RuntimeError(
        'pangocffi does not support Python 2.x. Please use Python 3.'
    )

setuptools.setup(
    setup_requires=['pytest-runner'],
    cffi_modules=[
        'pangocffi/ffi_build.py:ffi'
    ]
)
